<!DOCTYPE kpartgui SYSTEM "kpartgui.dtd">
<kpartgui name="zanshin" version="3">
 <MenuBar>
  <Menu name="file">
   <Action name="pages_project_add" />
   <Action name="pages_context_add" />
   <Action name="pages_remove" />
   <Separator />
   <Action name="page_view_add" />
   <Action name="page_view_remove" />
   <Action name="page_view_move" />
   <Action name="page_view_promote" />
  </Menu>
  <Menu name="go">
   <text>&amp;Go</text>
   <Action name="page_view_filter" />
   <Action name="page_view_future" />
   <Separator />
   <Action name="pages_go_previous" />
   <Action name="pages_go_next" />
   <Separator />
   <Action name="pages_go_to" />
  </Menu>
 </MenuBar>
 <ToolBar noMerge="1" name="mainToolBar" >
  <text context="@title:menu">Main Toolbar</text>
  <Action name="pages_go_previous" />
  <Action name="pages_go_next" />
  <Separator />
  <Action name="page_view_add" />
  <Separator />
  <Action name="page_view_filter" />
  <Action name="page_view_future" />
  <Separator />
  <Action name="page_run_task" />
 </ToolBar>
</kpartgui>
