#ifndef __MOCKITOPP_MATCHERS_HPP__
#define __MOCKITOPP_MATCHERS_HPP__

#include <mockitopp/matchers/base/any.hpp>
#include <mockitopp/matchers/base/equal.hpp>
#include <mockitopp/matchers/base/is_not.hpp>
#include <mockitopp/matchers/base/null.hpp>
#include <mockitopp/matchers/type/string_literal.hpp>

#endif //__MOCKITOPP_MATCHERS_HPP__
